@@ -3,7 +3,7 @@ package HTTP::Any::Curl;
 use strict;
 use warnings;
 
-use Net::Curl::Easy qw(/^CURLOPT_/ CURLE_OK CURLINFO_EFFECTIVE_URL CURLE_WRITE_ERROR CURLE_OPERATION_TIMEDOUT);
+use Net::Curl::Easy qw(/^CURLOPT_/ CURLE_OK CURLE_ABORTED_BY_CALLBACK CURLINFO_EFFECTIVE_URL);
 
 
 BEGIN {
@@ -24,25 +24,7 @@ sub do_http {
 		unless ($$opt{headers}{"Content-Type"}) {
 			push @headers, "Content-Type: application/x-www-form-urlencoded";
 		}
-		if (ref $$opt{body} eq "CODE") {
-			$easy->setopt(CURLOPT_POST, 1);
-			$easy->setopt(CURLOPT_UPLOAD, 1);
-			$easy->setopt(CURLOPT_CUSTOMREQUEST, "POST");
-
-			my $buf = "";
-			my $body_sub = $$opt{body};
-			$easy->setopt(CURLOPT_READFUNCTION, sub {
-				my ( $easy, $maxlen, $uservar ) = @_;
-				$buf ||= $body_sub->();
-				if ($buf) {
-					return \ substr $buf, 0, $maxlen, "";
-				} else {
-					return CURLE_OK;
-				}
-			} );
-		} else {
-			$easy->setopt(CURLOPT_POSTFIELDS, $$opt{body});
-		}
+		$easy->setopt(CURLOPT_POSTFIELDS, $$opt{body});
 	}
 
 	$easy->setopt(CURLOPT_HTTPHEADER, \@headers) if @headers;
@@ -63,7 +45,7 @@ sub do_http {
 
 	$easy->setopt(CURLOPT_WRITEHEADER, \ my $headers);
 
-	my $body = "";
+	my $body;
 	$easy->setopt(CURLOPT_FILE, \$body) unless $on_body;
 
 	$easy->setopt(CURLOPT_USERAGENT, $$opt{agent}) if $$opt{agent};
@@ -80,16 +62,20 @@ sub do_http {
 
 	$easy->setopt(CURLOPT_FORBID_REUSE, $$opt{persistent} ? 0 : 1) if exists $$opt{persistent};
 
-	my $max_size = $$opt{max_size};
-	my $aborted_by_max_size = 0;
+	if (my $max_size = $$opt{max_size}) {
+		$easy->setopt(CURLOPT_PROGRESSFUNCTION, sub {
+				my ($easy, $dltotal, $dlnow, $ultotal, $ulnow, $uservar) = @_;
+				if ($dlnow > $max_size) {
+					return 1;
+				}
+				return 0;
+			});
+		$easy->setopt(CURLOPT_NOPROGRESS, 0);
+	}
 
-	my $body_size = 0;
-
-	if ($max_size or $on_header or $on_body) {
+	if ($on_header or $on_body) {
 		my $cb_write = sub {
 			my ($easy, $data, $uservar) = @_;
-			my $size = length $data;
-			$body_size += $size;
 			if ($on_header) {
 				my ($is_success, $headers, $redirects) = headers($easy, $url, $headers);
 				my $r = $on_header->($is_success, $headers, $redirects);
@@ -98,14 +84,8 @@ sub do_http {
 			}
 			if ($on_body) {
 				$on_body->($data) or return 0;
-			} else {
-				$body .= $data;
 			}
-			if ($max_size and $body_size > $max_size) {
-				$aborted_by_max_size = 1;
-				return 0;
-			}
-			return $size;
+			return length $data;
 		};
 		$easy->setopt(CURLOPT_WRITEFUNCTION, $cb_write);
 	}
@@ -115,15 +95,7 @@ sub do_http {
 
 		if ($headers) {
 			my ($is_success, $headers, $redirects) = headers($easy, $url, $headers);
-			if ($result == CURLE_WRITE_ERROR and $aborted_by_max_size) {
-				$is_success = 0;
-				$$headers{"Status"} = 599;
-				$$headers{"Reason"} = "MaxSize";
-			} elsif ($result == CURLE_OPERATION_TIMEDOUT) {
-				$is_success = 0;
-				$$headers{"Status"} = 599;
-				$$headers{"Reason"} = "Timeout";
-			}
+			$$headers{'client-aborted'} = 'max_size' if $result == CURLE_ABORTED_BY_CALLBACK;
 			$cb->($is_success, $body, $headers, $redirects);
 		} else {
 			$cb->(0, undef, { Status => 500, Reason => $easy->error(), URL => $url }, []);
